"""An UpdateFrame holds cell-circuit and tessellation data."""

from typing import List
from qiskit import QuantumCircuit
from qiskit.circuit.quantumregister import Qubit, QuantumRegister
from .tessellation import Tessellation
from .exceptions import (CircuitWrongShapeForCell)


class UpdateFrame:
    """Create a large circuit from a tessellated small circuit."""

    cell_circuit: QuantumCircuit
    tessellation: Tessellation
    full_circuit_instructions: List

    def __init__(self,
                 tessellation: Tessellation,
                 qiskit_circuit: QuantumCircuit):
        """Hold the circuit to be applied to each cell in the tessellation.
        
        Note that you can construct a qiskit circuit via
        * qiskit.QuantumCircuit.from_qasm_str
        * qiskit.QuantumCircuit.from_qasm_file
        """        
        self.cell_circuit = qiskit_circuit

        self.tessellation = tessellation
        self.full_circuit_instructions = _wind_circuit_around_loop(
            self.cell_circuit, self.tessellation)

    def __str__(self):
        """Human-readable string representation."""
        return f"UpdateFrame(circuit: {self.cell_circuit} on each cell of {str(self.tessellation)})"


def _wind_circuit_around_loop(circuit: QuantumCircuit, tessellation: Tessellation):
    """Return the tessellated circuit as a list of instructions.

    Args:
        circuit (QuantumCircuit): Circuit defined on qubits in the first cell, to be tessellated to all cells.
        tessellation (Tessellation): Tessellation of the qubits into cells.

    Raises:
        CircuitWrongShapeForCell: The circuit cannot use more qubits than there are qubits in the first cell.

    Returns:
        List[circuit instructions]: List of instructions that will later be combined into a circuit.
    """
    if len(circuit.qubits) > len(tessellation.cells[0]):
        raise CircuitWrongShapeForCell(
            circuit.qubits, len(tessellation.cells[0]))
    next_column = []

    qreg = QuantumRegister(tessellation.size, circuit.qregs[0].name)

    for cell in tessellation.cells:
        for instruction, qargs, cargs in circuit.data:
            # Create a new instruction pointing at the correct qubits
            instruction_context = instruction, [
                Qubit(qreg, cell[q.index]) for q in qargs], cargs
            next_column.append(instruction_context)
    return next_column

"""
The MIT License (MIT)

Copyright (c) 2021 Hector Miller-Bakewell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


"""Exceptions for the PQCA module."""


class PQCAException(Exception):
    """Base exception class for the pqca module."""


class PartitionUnevenlyCoversQubits(PQCAException):
    """The partition must contain each qubit exactly once."""

    def __init__(self, cells):
        """Create PartitionUnevenlyCoversQubits exception."""
        super().__init__(
            f"Each qubit must appear exactly once in {cells}.")


class EmptyCellException(PQCAException):
    """Each cell must not be empty."""

    def __init__(self):
        """Create EmptyCellException exception."""
        super().__init__("Cells cannot be empty.")


class NoCellsException(PQCAException):
    """There must be at least one cell."""

    def __init__(self):
        """Create NoCellsException exception."""
        super().__init__("There must be at least one cell.")


class IrregularCoordinateDimensions(PQCAException):
    """Two coordinates must be of the same dimension to be compared."""

    def __init__(self, coordinate_1, coordinate_2):
        """Create IrregularCoordinateDimensions exception."""
        super().__init__(
            f"{coordinate_1} and {coordinate_2} " +
            "must be of the same length, and divide component-wise.")


class IrregularCellSize(PQCAException):
    """Each cell must be the same size."""

    def __init__(self, cells):
        """Create IrregularCellSize exception."""
        super().__init__(f"Not all cells in {cells} are of the same size.")


class CircuitWrongShapeForCell(PQCAException):
    """The circuit must be applied to qubits 0...(n-1)."""

    def __init__(self, qubits, size_cell):
        """Create CircuitWrongShapeForCell exception."""
        super().__init__(f"Could not apply circuit with qubits {qubits} \
            to cell of size {size_cell}")


class BackendError(PQCAException):
    """Pass backend errors through to user."""

"""
The MIT License (MIT)

Copyright (c) 2021 Hector Miller-Bakewell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


"""Partitioned Quantum Cellular Automaton."""

from typing import List, Callable
from qiskit import QuantumCircuit
import qiskit
from .update_frame import UpdateFrame


class Automaton:
    """Partitioned Quantum Cellular Automaton.

    Use `next(automaton)` to advance and return the new state.

    Has an internal state that is updated by applying the update circuit.
    The update circuit is formed by sequentially applying the update frames,
    and then measuring the qubits.

    The evaluation of the circuit is performed by the supplied backend.
    """

    def __init__(self, initial_state: List[int],
                 frames: List[UpdateFrame], backend: Callable[[QuantumCircuit], List[int]]):
        """Automaton with a given initial state, update frames, and backend.

        Use `next(automaton)` to advance and return the new state.

        Args:
            initial_state (List[int]): List of 0s and 1s indicating starting state.
            frames (List[UpdateFrame]): List of update frames to be applied in sequence as the update step.
            backend (Callable[[QuantumCircuit], List[int]]): A function that evaluates a
                quantum circuit once and returns the resulting list of classical bits.
        """
        self.frames = frames
        self.backend = backend
        frame_instructions = map(
            lambda f: f.full_circuit_instructions, self.frames)
        self.update_instruction = [instruction for
                                   instructions in frame_instructions for
                                   instruction in instructions]

        size = len(initial_state)

        self.state = initial_state

        self.update_circuit = QuantumCircuit(size)
        for instruction, qargs, cargs in self.update_instruction:
            self.update_circuit.append(instruction, qargs, cargs)

    @property
    def preparation_circuit(self) -> qiskit.QuantumCircuit:
        """Circuit for preparing a register of qubits into the current state."""
        return _pattern_preparation_circuit(self.state)

    @property
    def combined_circuit(self) -> qiskit.QuantumCircuit:
        """Combine preparation and update circuit."""
        return self.preparation_circuit + self.update_circuit

    def _tick(self) -> None:
        """Update the state without returning anything."""
        assert self.backend, "Backend not yet assigned"
        next_pattern = self.backend(self.combined_circuit)
        self.state = next_pattern

    def __next__(self) -> List[int]:
        """Yield the next state after applying the update circuit.

        Returns:
            List[int]: The new state of the Automaton
        """
        self._tick()
        return self.state

    def __str__(self) -> str:
        """Represent as string."""
        frame_string = f"[{','.join([str(frame) for frame in self.frames])}]"
        return f"PQCA(state={self.state}, frames={frame_string})"


def _pattern_preparation_circuit(pattern: List[int]) -> qiskit.QuantumCircuit:
    """Create a circuit that encodes the given classical state.

    Args:
        pattern (List[int]): A list of 0s and 1s.

    Returns:
        qiskit.QuantumCircuit: A circuit that encodes the given classical bits as
            the tensor product of 0 and 1 states.
    """
    circuit = QuantumCircuit(len(pattern))
    for (qubit, boolean_value) in list(enumerate(pattern)):
        if boolean_value:
            circuit.x(qubit)
    return circuit


"""
The MIT License (MIT)

Copyright (c) 2021 Hector Miller-Bakewell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
